#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/stage/T3SequentialStager.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                22.04.2021
# Last Modified Date:  13.07.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from time import time
from collections.abc import Generator, Iterable, Sequence
from ampel.view.T3Store import T3Store
from ampel.view.T3DocView import T3DocView
from ampel.view.SnapView import SnapView
from ampel.model.UnitModel import UnitModel
from ampel.content.T3Document import T3Document
from ampel.abstract.AbsT3ReviewUnit import AbsT3ReviewUnit, T
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.t3.stage.BaseViewGenerator import BaseViewGenerator, T3Send
from ampel.t3.stage.T3BaseStager import T3BaseStager
from ampel.mongo.update.MongoStockUpdater import MongoStockUpdater


class SimpleGenerator(BaseViewGenerator[T]):

	def __init__(self, unit: AbsT3ReviewUnit, views: Iterable[T], stock_updr: MongoStockUpdater) -> None:
		super().__init__(unit_name = unit.__class__.__name__, stock_updr = stock_updr)
		self.views = views

	def __iter__(self) -> Generator[T, T3Send, None]:
		l = self.stocks
		for v in self.views:
			l.append(v.id)
			yield v


class T3SequentialStager(T3BaseStager):
	"""
	A stager that calls t3 units 'process' methods sequentially.

	1) The source AmpelBuffer generator is fully consumed
	and buffers are converted to views which are stored in memory (all of them).
	Each T3 units is provided with a new generator based on those views.

	2) Results from upstream t3 units are made available for downstream units
	through the t3 store instance which is updated after each unit execution
	unless propagate is set to False
	"""

	propagate: bool = True

	#: t3 units (AbsT3ReviewUnit) to execute
	execute: Sequence[UnitModel]


	def __init__(self, **kwargs) -> None:

		super().__init__(**kwargs)

		if self.logger.verbose > 1:
			self.logger.debug(f"Setting up {self.__class__.__name__}")

		self.units = [
			self.get_unit(model, self.channel)
			for model in self.execute
		]


	def stage(self,
		gen: Generator[AmpelBuffer, None, None],
		t3s: T3Store
	) -> None | Generator[T3Document, None, None]:

		for t3_unit, views in self.get_views(gen).items():

			sg = SimpleGenerator(t3_unit, views, self.stock_updr)
			ts = time()
			if (ret := t3_unit.process(sg, t3s)):
				if (x := self.handle_t3_result(t3_unit, ret, t3s, sg.stocks, ts)):
					if self.propagate:
						t3s.add_view(
							T3DocView.of(x, self.context.config)
						)
					yield x


	def get_views(self, gen: Generator[AmpelBuffer, None, None]) -> dict[AbsT3ReviewUnit, list[SnapView]]:

		Views: set[type[SnapView]] = {u._View for u in self.units}
		conf = self.context.config

		if len(Views) == 1:
			View = next(iter(Views))
			if self.paranoia_level:
				buffers = list(gen)
				return {
					unit: [View.of(ab, conf) for ab in buffers]
					for unit in self.units
				}
			else:
				vs: list[SnapView] = [View.of(ab, conf) for ab in buffers]
				return {unit: vs for unit in self.units}
		else:

			buffers = list(gen)
			if self.paranoia_level:
				return {
					unit: [View.of(ab, conf) for ab in buffers]
					for unit, View in (lambda x: [(u, u._View) for u in x])(self.units)
				}

			else:

				optd: dict[type[SnapView], list[AbsT3ReviewUnit]] = {}
				for unit in self.units:
					if unit._View not in optd:
						optd[unit._View] = []
					optd[unit._View].append(unit)

				d = {}
				for View, units in optd.items():
					vs = [View.of(ab, conf) for ab in buffers]
					for unit in units:
						d[unit] = vs
				return d



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/collector/T02ConfigCollector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                28.10.2019
# Last Modified Date:  23.05.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import hashlib, json
from typing import Any
from ampel.util.hash import build_unsafe_dict_id
from ampel.util.mappings import dictify
from ampel.config.collector.AbsDictConfigCollector import AbsDictConfigCollector
from ampel.log import VERBOSE


class T02ConfigCollector(AbsDictConfigCollector):
	"""
	Collect and hash T2 configurations contained in T0 processes (hence the name T02)
	Note: uses hash algorithm from ampel.util.hash.build_unsafe_dict_id.
	An additional hash collision test is performed using SHA256
	"""

	def add(self,
		arg: dict[str, Any],
		dist_name: str,
		version: str | float | int,
		register_file: str
	) -> int:

		arg = dictify(arg)

		# Uses xxhash algorithm
		int_hash = build_unsafe_dict_id(arg)

		if self.get(int_hash):

			# Build SHA256 hash
			hash1 = hashlib.sha256(bytes(json.dumps(self.get(int_hash), sort_keys=True), "utf-8")).digest()
			hash2 = hashlib.sha256(bytes(json.dumps(arg, sort_keys=True), "utf-8")).digest()

			if hash1 != hash2:
				raise ValueError("Hash collision detected")

			if self.verbose:
				self.logger.log(VERBOSE, f"Re-using T2 config hash: {int_hash}")

			return int_hash

		if self.verbose:
			self.logger.log(VERBOSE, f"Adding T2 config hash: {int_hash}")

		self.__setitem__(int_hash, arg)
		return int_hash



from ampel.model.UnitModel import UnitModel
import pytest

from ampel.model.DPSelection import DPSelection


@pytest.mark.parametrize(
    "filter,count",
    [
        (None, 0),
        ("SimpleTagFilter", 5),
        (UnitModel(unit="SimpleTagFilter", config={"require": ["good"]}), 2),
        (UnitModel(unit="SimpleTagFilter", config={"forbid": ["bad"]}), 3),
        (
            UnitModel(
                unit="SimpleTagFilter", config={"require": ["good"], "forbid": ["bad"]}
            ),
            1,
        ),
    ],
    ids=["none", "noop", "require", "forbid", "both"],
)
def test_SimpleTagFilter(mock_context, filter, count):
    dp = DPSelection(filter=filter)
    filterer, sorter, slicer = dp.tools()
    if filter is None:
        assert filterer is None
    else:
        datapoints = [
            {"id": -2},
            {"id": -1, "tag": []},
            {"id": 0, "tag": ["good"]},
            {"id": 1, "tag": ["bad"]},
            {"id": 2, "tag": ["bad", "good"]},
        ]

        assert len(filterer.apply(datapoints)) == count



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/util/debug.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                24.07.2022
# Last Modified Date:  17.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import sys, pdb, traceback

"""
Usage:
from ampel.util.debug import mp_breakpoint
-> mp_breakpoint()

from ampel.util.debug import trace_prints
-> trace_prints()
"""

# https://stackoverflow.com/questions/26289153/how-to-use-ipdb-set-trace-in-a-forked-process
class ForkedPdb(pdb.Pdb):
	"""
	A Pdb subclass that may be used within a multiprocessing process

	Forword: If you are not using multiprocessing, just use breakpoint()

	Usage:
	insert: "ForkedPdb().set_trace()" at the wished position
	<run your job>
	a pdb shell should pop up, type "interact" and press enter
	<do debug>
	when you are done, press CTRL-D
	you are back in the pdb shell, type "continue" and press enter
	Hope this helps
	"""

	def interaction(self, *args, **kwargs):
		_stdin = sys.stdin
		try:
			sys.stdin = open('/dev/stdin')
			pdb.Pdb.interaction(self, *args, **kwargs)
		finally:
			sys.stdin = _stdin


# https://stackoverflow.com/questions/1617494/finding-a-print-statement-in-python
class TracePrints:

	def __init__(self):    
		self.stdout = sys.stdout

	def write(self, s):
		self.stdout.write("Writing %r\n" % s)
		traceback.print_stack(file=self.stdout)


def trace_prints():
	original_stdout.flush()
	sys.stdout = TracePrints()

def stop_trace_prints():
	original_stdout.flush()
	sys.stdout = original_stdout

mp_breakpoint = ForkedPdb().set_trace
original_stdout = sys.stdout



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/update/MongoStockUpdater.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                15.10.2018
# Last Modified Date:  03.09.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from time import time
from bson import ObjectId
from pymongo.errors import BulkWriteError
from pymongo.operations import UpdateMany, UpdateOne
from typing import Any, Literal, get_args
from collections.abc import Sequence

from ampel.core.AmpelDB import AmpelDB
from ampel.types import ChannelId, Tag, StockId
from ampel.log import AmpelLogger, VERBOSE
from ampel.log.utils import report_exception
from ampel.mongo.utils import maybe_use_each
from ampel.enum.JournalActionCode import JournalActionCode
from ampel.struct.JournalAttributes import JournalAttributes
from ampel.content.JournalRecord import JournalRecord


tag_type = get_args(Tag) # type: ignore[misc]
chan_type = get_args(ChannelId) # type: ignore[misc]


class MongoStockUpdater:

	def __init__(self,
		ampel_db: AmpelDB, tier: Literal[-1, 0, 1, 2, 3], run_id: int,
		process_name: str, logger: AmpelLogger, raise_exc: bool = False,
		bump_updated: bool = True, update_journal: bool = True,
		extra_tag: None | Tag | Sequence[Tag] = None,
		auto_flush: int = 0
	) -> None:
		"""
		:param raise_exc: raise exception rather than populating 'troubles' collection
		"""

		self._ampel_db = ampel_db
		self.col_stock = ampel_db.get_collection('stock')
		self.run_id = run_id
		self.tier = tier
		self.raise_exc = raise_exc
		self.update_journal = update_journal
		self.bump_updated = bump_updated
		self.auto_flush = auto_flush
		self.process_name = process_name
		self.extra_tag = extra_tag
		self.logger = logger
		self.reset()


	def reset(self) -> None:
		self._updates: list[UpdateOne | UpdateMany] = []
		self._one_updates: dict[StockId, UpdateOne] = {}
		self._multi_updates: dict[StockId, list[UpdateMany]] = {}


	def new_journal_record(self,
		unit: None | int | str = None,
		channels: None | ChannelId | Sequence[ChannelId] = None,
		action_code: None | JournalActionCode = None,
		doc_id: None | ObjectId = None,
		now: None | int | float = None
	) -> JournalRecord:

		ret: JournalRecord = {
			'tier': self.tier,
			'ts': now if now else int(time()),
			'process': self.process_name,
			'run': self.run_id,
			'action': action_code or JournalActionCode(0)
		}

		if channels:
			ret['channel'] = channels

		if unit:
			ret['unit'] = unit

		if doc_id:
			ret['doc'] = doc_id.binary

		if self.extra_tag:
			ret['tag'] = self.extra_tag

		return ret


	def add_journal_record(self,
		stock: StockId | Sequence[StockId],
		jattrs: None | JournalAttributes = None,
		tag: None | Tag | Sequence[Tag] = None,
		name: None | str | Sequence[str] = None,
		trace_id: None | dict[str, int] = None,
		action_code: None | JournalActionCode = None,
		doc_id: None | ObjectId = None,
		unit: None | int | str = None,
		channel: None | ChannelId | Sequence[ChannelId] = None,
		now: None | int | float = None
	) -> JournalRecord:
		"""
		:returns: the JournalRecord dict instance associated with the stock document(s) update
		is returned by this method so that customization can be made if necessary.
		Note that these customizations must be made before additional operations such as add_tag.
		Note that the associated update operation does more than just adding a journal record,
		it also modifies the "updated" field of stock document(s).
		"""

		jrec = self.new_journal_record(unit, channel, action_code, doc_id, now)

		if jattrs:

			if jattrs.tag:
				self.include_jtags(jrec, jattrs.tag)

			if jattrs.extra:
				jrec['extra'] = jattrs.extra

			if jattrs.code and jattrs.code > 0:
				jrec['code'] = jattrs.code

		if trace_id:
			jrec['traceid'] = trace_id

		upd: dict[str, Any] = {'$push': {'journal': jrec}}

		if tag or name:

			upd['$addToSet'] = (
				({'tag': tag if isinstance(tag, (str, int)) else maybe_use_each(tag)} if tag else {}) |
				({'name': name if isinstance(name, str) else maybe_use_each(name)} if name else {})
			)

		if self.bump_updated:

			upd['$max'] = {'ts.any.upd': jrec['ts']}

			if channel:
				if isinstance(channel, chan_type):
					upd['$max'][f'ts.{channel}.upd'] = jrec['ts']
				else:
					upd['$max'].update({
						f'ts.{chan}.upd': jrec['ts']
						for chan in channel # type: ignore[union-attr]
					})

		if self.update_journal:

			# Current strategy:
			# - case: UpdateMany + UpdateMany: do nothing or put differently: submit the two ops as is
			# - case: UpdateOne + UpdateMany: update op UpdateOne and no longer match target stock in UpdateMany
			# - case: UpdateOne + UpdateOne: merge updates
			if not isinstance(stock, (int, str)):
				self._add_many_update(list(stock), upd)
			else:
				self._add_one_update(stock, upd)

		return jrec


	def add_name(self, stock: StockId, name: str | Sequence[str]) -> None:
		self._add_one_update(
			stock,
			{'$addToSet': {'name': name if isinstance(name, str) else {'$each': name}}}
		)


	def add_tag(self,
		stock: StockId | Sequence[StockId],
		tag: Tag | Sequence[Tag]
	) -> None:

		upd = {'$addToSet': {'tag': tag if isinstance(tag, tag_type) else {'$each': tag}}}

		if isinstance(stock, (int, bytes, str)): # StockId
			self._add_one_update(stock, upd)
		else: # Sequence[StockId]
			self._add_many_update(stock if isinstance(stock, list) else list(stock), upd)


	def _add_one_update(self, stock: StockId, upd: dict[str, Any]) -> None:

		uo = UpdateOne({'stock': stock}, upd)

		if stock in self._multi_updates:
			for um in self._multi_updates.pop(stock):
				um._filter['stock']['$in'].remove(stock)
				self._merge_updates(uo, um._doc)

		if stock in self._one_updates:
			self._merge_updates(self._one_updates[stock], upd)
		else:
			self._one_updates[stock] = uo
			self._updates.append(uo)

			if self.auto_flush and len(self._updates) > self.auto_flush:
				self.flush()


	def _add_many_update(self, stocks: list[StockId], upd: dict[str, Any]) -> None:

		um = UpdateMany({'stock': {'$in': stocks}}, upd)

		for s in iter(stocks):
			if s in self._one_updates:
				self._merge_updates(self._one_updates[s], upd)
				stocks.remove(s)

		if stocks:

			for s in stocks:
				if s in self._multi_updates:
					self._multi_updates[s].append(um)
				else:
					self._multi_updates[s] = [um]

			self._updates.append(um)
			if self.auto_flush and len(self._updates) > self.auto_flush:
				self.flush()


	def _merge_updates(self, op: UpdateOne, d: dict) -> None:
		"""
		modifies provided UpdateOne structure
		:raises: ValueError in case update structures are not conform ex: {'$addToSet': {'name': {'a': 1}}}
		"""

		opd = op._doc

		if '$max' in d:
			if '$max' in opd:
				for k in d['$max']:
					if k in opd['$max']:
						opd['$max'][k] = max(opd['$max'][k], d['$max'][k])
					else:
						opd['$max'][k] = d['$max'][k]
			else:
				opd['$max'] = d['$max']

		if '$set' in d:
			if '$set' in opd:
				for k in d['$set']:
					# Second update prevails as would occur if there was two ops
					opd['$set'][k] = d['$set'][k]
			else:
				opd['$set'] = d['$set']


		for mop in ('$push', '$addToSet'):
			if mop in d:
				if mop in opd:
					for k in d[mop]:
						if k in opd[mop]:
							if isinstance(opd[mop][k], str) or (isinstance(opd[mop][k], dict) and '$each' not in opd[mop][k]):
								opd[mop][k] = {'$each': [opd[mop][k]]}
							if isinstance(d[mop][k], str):
								opd[mop][k]['$each'].append(d[mop][k])
							elif isinstance(d[mop][k], dict):
								opd[mop][k]['$each'].append(d[mop][k]['$each'] if '$each' in d[mop][k] else d[mop][k])
							else:
								raise ValueError(f"Unrecognized d[{mop}][{k}] value: {d[mop][k]}")
						else:
							opd[mop][k] = d[mop][k]
				else:
					opd[mop] = d[mop]

		for k in d.keys() - ['$set', '$max', '$push', '$addToSet']:
			opd[k] = d[k]


	def flush(self) -> None:

		if not self._updates:
			return

		jupds = self._updates
		self.reset()

		try:

			if self.logger.verbose > 1:
				for el in jupds:
					self.logger.log(VERBOSE, f"Journal update: {str(el)}")

			self.col_stock.bulk_write(jupds)

			if self.logger.verbose:
				self.logger.log(VERBOSE,
					f"{len(jupds)} journal entr{'ies' if len(jupds) > 1 else 'y'} inserted"
				)

		except Exception as e:

			if self.raise_exc:
				if isinstance(e, BulkWriteError):
					raise ValueError(f"Journal update error: {e.details}")
				raise ValueError("Journal update error")

			info: dict[str, Any] = {
				'process': self.process_name,
				'msg': 'Exception in flush()',
				'journalUpdateCount': len(jupds)
			}

			if isinstance(e, BulkWriteError):
				info['BulkWriteError'] = str(e.details)

			# Populate troubles collection
			report_exception(self._ampel_db, self.logger, exc=e, info=info)


	@staticmethod
	def include_jtags(jrec: JournalRecord, tag: Tag | Sequence[Tag]):
		""" Modifies the input JournalRecord dict """

		if tag:

			# journal record also contains tag(s)
			if 'tag' in jrec:

				# tweak request is about a single tag (not a sequence of tags)
				if isinstance(tag, tag_type):
					# journal record contains single tag (not a sequence of tags)
					if isinstance(jrec['tag'], tag_type):
						jrec['tag'] = [jrec['tag'], tag] # type: ignore[list-item]
					else:
						jrec['tag'].append(tag) # type: ignore[union-attr]

				# multi-tag tweak request
				else:
					# journal record contains single tag (not a sequence of tags)
					if isinstance(jrec['tag'], tag_type):
						jrec['tag'] = [*jrec['tag'], tag] # type: ignore[list-item, misc]
					else:
						jrec['tag'] = jrec['tag'] + tag # type: ignore[operator]

			# journal record contains no tag
			else:
				jrec['tag'] = tag



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/supply/select/T3StockSelector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                06.12.2019
# Last Modified Date:  16.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from pymongo.cursor import Cursor
from typing import Literal, Any
from ampel.types import ChannelId, Tag
from ampel.mongo.query.stock import build_stock_query
from ampel.util.logicschema import to_logical_dict
from ampel.log.utils import safe_query_dict
from ampel.log.AmpelLogger import AmpelLogger, VERBOSE
from ampel.abstract.AbsT3Selector import AbsT3Selector
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.OneOf import OneOf
from ampel.model.time.TimeConstraintModel import TimeConstraintModel
from ampel.util.collections import try_reduce


class T3StockSelector(AbsT3Selector):
	"""
	Default stock selector used by T3Processor
	
	Example configuration::
	  
	  {
	     "created": {"after": {"use": "$timeDelta", "arguments": {"days": -40}}},
	     "updated": {"after": {"use": "$timeDelta", "arguments": {"days": -1}}},
	     "channel": "HU_GP_CLEAN",
	     "tags": {"with": "ZTF", "without": "HAS_ERROR"}
	  }
	"""

	#: Select by creation time
	created: None | TimeConstraintModel = None

	#: Select by modification time
	updated: None | TimeConstraintModel = None

	#: Select by channel
	channel: None | ChannelId | AnyOf[ChannelId] | AllOf[ChannelId] | OneOf[ChannelId] = None

	#: Select by tag
	tag: None | dict[Literal['with', 'without'], Tag | AllOf[Tag] | AnyOf[Tag] | OneOf[Tag] | dict] = None

	#: Custom selection (ex: {'run': {'$gt': 10}})
	custom: None | dict[str, Any] = None


	def __init__(self, logger: AmpelLogger, **kwargs):

		if 'channel' in kwargs:
			kwargs['channel'] = to_logical_dict(kwargs['channel'], 'channel')

		if 'tag' in kwargs:
			kwargs['tag'] = {
				k: to_logical_dict(v, 'tag')
				for k, v in kwargs['tag'].items()
			}

		if logger is None:
			raise ValueError("Parameter logger cannot be None")

		self.logger = logger
		super().__init__(**kwargs)


	# Override/Implement
	def fetch(self) -> None | Cursor:

		# Build query for matching transients using criteria defined in config
		match_query = build_stock_query(
			channel = self.channel,
			tag = self.tag,
			time_created = self.created.get_query_model(db=self.context.db)
				if self.created else None,
			time_updated = self.updated.get_query_model(db=self.context.db)
				if self.updated else None,
		)

		if self.custom:
			match_query.update(self.custom)

		col = self.context.db.get_collection('stock')

		if self.logger.verbose:
			self.logger.log(
				VERBOSE,
				f"Executing stock search query [{col.database.name} "
				f"{try_reduce(list(col.database.client.nodes))}]",
				extra=safe_query_dict(match_query)
			)

		# Execute 'find transients' query
		return col.find(match_query, {'stock': 1})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                ampel/model/builder/BuilderAliasModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                09.10.2019
# Last Modified Date:  11.10.2019
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.base.AmpelBaseModel import AmpelBaseModel

class BuilderAliasModel(AmpelBaseModel):

	t0: None | dict[str, Any]
	t1: None | dict[str, Any]
	t2: None | dict[str, Any]
	t3: None | dict[str, Any]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsT3ControlUnit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                08.12.2021
# Last Modified Date:  17.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Generator
from ampel.types import Traceless, OneOrMany, ChannelId
from ampel.view.T3Store import T3Store
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.core.ContextUnit import ContextUnit
from ampel.core.EventHandler import EventHandler
from ampel.content.T3Document import T3Document
from ampel.log.AmpelLogger import AmpelLogger


class AbsT3ControlUnit(ContextUnit, AmpelABC, abstract=True):
	"""
	Abstract class for control T3 units, which like AbsT3PlainUnit but
	unlike AbsT3ReviewUnit, receive just a T3Store instance via
	the method process()
	"""

	logger: Traceless[AmpelLogger]
	event_hdlr: Traceless[EventHandler]
	channel: None | OneOrMany[ChannelId] = None


	@abstractmethod
	def process(self, t3s: T3Store) -> None | Generator[T3Document, None, None]:
		"""
		The content of the t3 store is dependent on:
		- the configuration of the 'include' option of the underlying t3 process
		- previously run t3 units if the option 'propagate' is activated
		"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/collector/ForwardProcessConfigCollector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                03.03.2020
# Last Modified Date:  04.03.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from collections.abc import Sequence
from ampel.config.collector.ConfigCollector import ConfigCollector
from ampel.config.collector.AbsForwardConfigCollector import AbsForwardConfigCollector
from ampel.log import VERBOSE


class ForwardProcessConfigCollector(AbsForwardConfigCollector):

	def get_path(self, # type: ignore
		arg: dict[str, Any],
		dist_name: str,
		version: str | float | int,
		register_file: str,
	) -> None | Sequence[int | str]:

		if not isinstance(arg, dict) or 'tier' not in arg:
			self.error(
				"Unsupported process definition" +
				ConfigCollector.distrib_hint(dist_name, register_file)
			)
			return None

		path = ["process", "ops" if arg["tier"] is None else f"t{arg['tier']}"]

		if self.verbose:
			self.logger.log(VERBOSE,
				f"Routing process '{arg['name']}' to '{'.'.join(path)}'"
			)

		return path



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/ingest/T1Combine.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                10.03.2020
# Last Modified Date:  27.05.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Sequence
from ampel.model.UnitModel import UnitModel
from ampel.model.ingest.T2Compute import T2Compute

class T1Combine(UnitModel[str]):
	"""
	Combine datapoints into t1 document and create associated t2 documents
	"""

	#: Filter result codes that should trigger this T1.
	#: If not specified, T1 documents will be created in response to any passing alert.
	group: None | int | Sequence[int] = None

	#: Create or update :class:`T2 documents <ampel.content.T2Document.T2Document>`
	#: bound to :class:`compounds <ampel.content.T1Document.T1Document>`
	state_t2: None | Sequence[T2Compute] = []

	#: Create or update :class:`T2 documents <ampel.content.T2Document.T2Document>`
	#: bound to :class:`datapoints <ampel.content.DataPoint.DataPoint>`
	#: based on the t1 combine unit result
	point_t2: None | Sequence[T2Compute] = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/time/TimeStringModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                29.09.2018
# Last Modified Date:  06.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Literal
from datetime import datetime
from ampel.base.AmpelBaseModel import AmpelBaseModel


class TimeStringModel(AmpelBaseModel):

	match_type: Literal['time_string']
	dateTimeStr: str
	dateTimeFormat: str

	def get_timestamp(self, **kwargs) -> float:

		return datetime \
			.strptime(self.dateTimeStr, self.dateTimeFormat) \
			.timestamp()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/job/utils.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                Unspecified
# Last Modified Date:  13.08.2022
# Last Modified By:    jvs

import re
from typing import Union, Any, Callable
from ampel.model.job.ExpandWithItems import ExpandWithItems
from ampel.model.job.ExpandWithSequence import ExpandWithSequence
from ampel.util.recursion import walk_and_process_dict

ExpandWith = Union[None, ExpandWithItems, ExpandWithSequence]

def _parse_multiplier(values: dict[str, Any]) -> dict:

	if not isinstance(multiplier := values.pop("multiplier", 1), int):
		raise TypeError("multiplier must be an int")

	if multiplier > 1:
		assert (
			"expand_with" not in values
		), "multiplier and expand_with may not be used together"
		values |= {"expand_with": {"sequence": {"count": multiplier}}}

	return values

def _transform_item(v: str, transform: Callable[[str], str]) -> str:
	chunks = []
	pos = 0
	for match in re.finditer(r"\{\{(.*?)\}\}", v):
		if match.span()[0] > pos:
			chunks.append(v[pos: match.span()[0]])
		chunks.append(transform(match.groups()[0].strip()))
		pos = match.span()[1]
	if pos < len(v):
		chunks.append(v[pos: len(v)])
	return "".join(chunks)


def _transform_expressions_callback(path, k, d, **kwargs) -> None:
	for k, v in d.items():
		if isinstance(v, str):
			d[k] = _transform_item(v, kwargs["transform"])
		elif isinstance(v, list):
			d[k] = [
				_transform_item(vv, kwargs["transform"])
				if isinstance(vv, str)
				else vv
				for vv in v
			]


def transform_expressions(
	task_dict: dict, transformation: Callable[[str], str]
) -> dict:
	"""
	Replace any expressions of the form {{ expr }} with the result of
	transformation(expr)
	"""
	walk_and_process_dict(
		task_dict,
		callback = _transform_expressions_callback,
		transform = transformation
	)
	return task_dict



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/job/ExpandWithSequence.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                Unspecified
# Last Modified Date:  13.08.2022
# Last Modified By:    jvs

from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.model.job.SequenceWithEnd import SequenceWithEnd
from ampel.model.job.SequenceWithCount import SequenceWithCount

class ExpandWithSequence(AmpelBaseModel):

    sequence: SequenceWithCount | SequenceWithEnd

    def items(self):
        for i in self.sequence.items():
            yield self.sequence.format % i

    def __iter__(self):
        return self.items()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/purge/PurgeContentModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                18.06.2020
# Last Modified Date:  18.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Literal
from ampel.base.AmpelBaseModel import AmpelBaseModel


class PurgeContentModel(AmpelBaseModel):
	"""
	:param delay: time of inactivity in days after which entities will be purged
	:param format: output format exported by ampel.
	- bson: (native format) use this format if you plan on working on archives with your own mongodb and ampel tools
	- json: ampel ids and tags will be converted from hashes back to strings
	:param unify: whether to merge documents from all tiers into one unique document for each stock id
	This can facilitate further analysis. Note that because BSON document cannot exceed 16MB,
	this parameter can only by used in combination with the 'json' format.
	:param compress: whether to compress the output file

	Note: For now, we do not offer the option to include logs into the 'merged' structure.
	The purge of logs and content is done with two different, independant processes.
	Feel free to implement your own UnifiedPurger and associated UnifiedPurgeModel if you need to.
	"""
	delay: int
	format: Literal['bson', 'json']
	unify: bool = False
	compress: bool = True

	def __init__(self, **kwargs) -> None:
		super().__init__(**kwargs) # type: ignore[call-arg]
		if self.format == 'bson' and self.unify:
			raise ValueError(
				"Parameter 'unify' can only by used in combination with " +
				"the 'json' format (please see docstring)"
			)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/update/MongoT3Ingester.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                30.05.2021
# Last Modified Date:  30.05.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from pymongo import UpdateOne
from typing import Any
from ampel.enum.DocumentCode import DocumentCode
from ampel.content.T3Document import T3Document
from ampel.mongo.utils import maybe_use_each
from ampel.abstract.AbsDocIngester import AbsDocIngester


class MongoT3Ingester(AbsDocIngester[T3Document]):

	def ingest(self, doc: T3Document) -> None:
		raise NotImplementedError("Soon...")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/T3ReviewUnitExecutor.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                12.12.2021
# Last Modified Date:  14.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Annotated
from collections.abc import Generator
from ampel.types import Traceless, ChannelId
from ampel.view.T3Store import T3Store
from ampel.abstract.AbsT3ControlUnit import AbsT3ControlUnit
from ampel.t3.T3DocBuilder import T3DocBuilder
from ampel.content.T3Document import T3Document
from ampel.model.UnitModel import UnitModel
from ampel.model.t3.T3DocBuilderModel import T3DocBuilderModel
from ampel.abstract.AbsT3Supplier import AbsT3Supplier
from ampel.abstract.AbsT3Stager import AbsT3Stager
from ampel.log.AmpelLogger import AmpelLogger


class T3ReviewUnitExecutor(AbsT3ControlUnit, T3DocBuilder):

	logger: Traceless[AmpelLogger]

	# Require single channel for now (super classes allow multi-channel)
	channel: None | ChannelId = None

	#: Unit must be a subclass of AbsT3Supplier
	supply: Annotated[UnitModel, AbsT3Supplier]

	#: Unit must be a subclass of AbsT3Stager
	stage: Annotated[UnitModel, AbsT3Stager]


	def process(self, t3s: T3Store) -> None | Generator[T3Document, None, None]:

		try:

			supplier = self.context.loader.new_context_unit(
				model = self.supply,
				context = self.context,
				sub_type = AbsT3Supplier,
				logger = self.logger,
				event_hdlr = self.event_hdlr
			)

			# Stager unit
			#############

			if self.stage.unit not in self.context.config._config['unit']:
				raise ValueError(f"Unknown stager unit: {self.stage.unit}")

			if "T3DocBuilder" in self.context.config._config['unit'][self.stage.unit]['base']:
				kwargs = {
					k: getattr(self, k)
					for k in T3DocBuilderModel.get_model_keys()
					if hasattr(self, k) and not (k in self._defaults and getattr(self, k) == self._defaults[k])
				}
			else:
				kwargs = {}

			stager = self.context.loader.new_context_unit(
				model = self.stage,
				context = self.context,
				sub_type = AbsT3Stager,
				logger = self.logger,
				event_hdlr = self.event_hdlr,
				channel = (
					self.stage.config['channel'] # type: ignore
					if self.stage.config and self.stage.config.get('channel') # type: ignore[union-attr]
					else self.channel
				),
				**kwargs
			)

			self.logger.info("Running stager", extra={'unit': self.stage.unit})
			return stager.stage(supplier.supply(t3s), t3s)

		except Exception as e:
			self.event_hdlr.handle_error(e, self.logger)

		return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/builder/DistConfigBuilder.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                09.10.2019
# Last Modified Date:  14.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import json, yaml, pkg_resources, os, re
from functools import partial
from pkg_resources import EggInfoDistribution, DistInfoDistribution # type: ignore[attr-defined]
from typing import Any
from collections.abc import Callable
from ampel.config.builder.ConfigBuilder import ConfigBuilder
from ampel.util.distrib import get_dist_names, get_files
from ampel.log import VERBOSE, SHOUT


class DistConfigBuilder(ConfigBuilder):
	"""
	Subclass of ConfigLoader allowing to automatically detect and load
	various configuration files contained in ampel sub-packages installed with pip
	"""

	def load_distributions(self,
		prefixes: list[str] = ["pyampel-", "ampel-"],
		conf_dirs: list[str] = ["conf"],
		exts: list[str] = ["json", "yaml", "yml"]
	) -> None:
		"""
		:param prefixes: loads all known conf files from all distributions with name starting with prefixes
		:param conf_dirs: loads only conf files from these directories
		:param exts: loads only conf files with these extensions
		"""
		dist_names = get_dist_names(prefixes)

		if dist_names:
			self.logger.log(SHOUT, f"Detected ampel components: '{dist_names}'")

		for dist_name in dist_names:

			self.logger.break_aggregation()
			s = f"Checking distribution '{dist_name}'"
			self.logger.log(VERBOSE, s)
			self.logger.log(VERBOSE, "="*len(s))

			for conf_dir in conf_dirs:
				for ext in exts:
					self.load_distrib(dist_name, conf_dir, ext)

			if self.verbose:
				self.logger.log(VERBOSE, f"Done checking distribution '{dist_name}'")

		if self.verbose:
			self.logger.log(VERBOSE, "Done loading distributions")


	def load_distrib(self, dist_name: str, conf_dir: str = "conf", ext: str = "json") -> None:
		"""
		Loads all known conf files of the provided distribution (name)
		"""
		try:

			distrib = pkg_resources.get_distribution(dist_name)
			all_conf_files = get_files(dist_name, conf_dir, re.compile(f".*\.{ext}$")) # noqa

			if all_conf_files and self.verbose:
				self.logger.log(VERBOSE, "Following conf files will be parsed:")
				for el in all_conf_files:
					self.logger.log(VERBOSE, el)

			if ampel_conf := self.get_conf_file(all_conf_files, f"ampel.{ext}"):
				self.load_conf_using_func(distrib, ampel_conf, self.load_ampel_conf) # type: ignore

			# Channel, mongo (and template) can be defined by multiple files
			# in a directory named after the corresponding config section name
			for key in ("channel", "mongo", "process"):
				if specialized_conf_files := self.get_conf_files(all_conf_files, f"/{key}/"):
					for f in specialized_conf_files:
						self.load_conf_using_func(distrib, f, self.first_pass_config[key].add)

			# ("channel", "mongo", "resource")
			for key in self.first_pass_config.conf_keys.keys():
				if key == "alias":
					continue
				if section_conf_file := self.get_conf_file(all_conf_files, f"{key}.{ext}"):
					self.load_conf_using_func(distrib, section_conf_file, self.first_pass_config[key].add) # type: ignore

			# ("controller", "processor", "unit", "alias", "process")
			for unit_type in ("alias", "process"):
				if tier_conf_file := self.get_conf_file(all_conf_files, f"{unit_type}.{ext}"):
					self.load_conf_using_func(distrib, tier_conf_file, partial(self.register_tier_conf, unit_type))

			# Try to load templates from folder template (defined by 'Ampel-ZTF' for ex.)
			if template_conf_files := self.get_conf_files(all_conf_files, "/template/"):
				for f in template_conf_files:
					self.load_conf_using_func(distrib, f, self.register_channel_templates)

			# Try to load templates from template.conf
			if template_conf := self.get_conf_file(all_conf_files, "/template.{ext}"):
				self.load_conf_using_func(distrib, template_conf, self.register_channel_templates) # type: ignore

			if all_conf_files:
				self.logger.info(f"Not all conf files were loaded from distribution '{distrib.project_name}'")
				self.logger.info(f"Unprocessed conf files: {all_conf_files}")

		except Exception as e:
			self.error = True
			self.logger.error(
				f"Error occured while loading configuration files from the distribution '{dist_name}'",
				exc_info=e
			)


	def register_tier_conf(self,
		root_key: str,
		d: dict[str, Any],
		dist_name: str,
		version: str,
		file_rel_path: str,
	):
		for k in ("t0", "t1", "t2", "t3", "ops"):
			if k in d:
				self.first_pass_config[root_key][k].add(
					d[k],
					dist_name = dist_name,
					version = version,
					register_file = file_rel_path,
				)


	def load_conf_using_func(self,
		distrib: EggInfoDistribution | DistInfoDistribution,
		file_rel_path: str,
		func: Callable[[dict[str, Any], str, str, str], None]
	) -> None:

		try:

			if self.verbose:
				self.logger.log(VERBOSE,
					f"Loading {file_rel_path} from distribution '{distrib.project_name}'"
				)

			if file_rel_path.endswith("json"):
				load = json.loads
			elif file_rel_path.endswith("yml") or file_rel_path.endswith("yaml"):
				load = yaml.safe_load # type: ignore
			
			if os.path.isabs(file_rel_path):
				with open(file_rel_path) as f:
					payload = f.read()
			else:
				payload = distrib.get_resource_string(__name__, file_rel_path)

			func(
				load(payload),
				distrib.project_name,
				distrib.version,
				file_rel_path
			)

		except FileNotFoundError:
			self.error = True
			self.logger.error(
				f"File '{file_rel_path}' not found in distribution '{distrib.project_name}'"
			)

		except Exception as e:
			self.error = True
			self.logger.error(
				f"Error occured loading '{file_rel_path}' from distribution '{distrib.project_name}'",
				exc_info=e
			)


	def load_tier_config_file(self,
		distrib: EggInfoDistribution | DistInfoDistribution,
		file_rel_path: str,
		root_key: str
	) -> None:
		"""
		Files loaded by this method must have the following JSON structure:
		{
			"t0": { ... },
			"t1": { ... },
			...
		}
		whereby the t0, t1, t2 and t3 keys are optional (at least one is required though)
		"""

		try:

			if self.verbose:
				self.logger.log(VERBOSE,
					f"Loading {file_rel_path} from distribution '{distrib.project_name}'"
				)

			if file_rel_path.endswith("json"):
				load = json.loads
			elif file_rel_path.endswith("yml") or file_rel_path.endswith("yaml"):
				load = yaml.safe_load # type: ignore

			d = load(
				distrib.get_resource_string(__name__, file_rel_path)
				if not os.path.isabs(file_rel_path)
				else file_rel_path
			)

			for k in ("t0", "t1", "t2", "t3", "ops"):
				if k in d:
					self.first_pass_config[root_key][k].add(
						d[k],
						dist_name = distrib.project_name,
						version = distrib.version,
						register_file = file_rel_path,
					)

		except FileNotFoundError:
			self.error = True
			self.logger.error(
				f"File '{file_rel_path}' not found in distribution '{distrib.project_name}'"
			)

		except Exception as e:
			self.error = True
			self.logger.error(
				f"Error occured loading '{file_rel_path}' "
				f"from distribution '{distrib.project_name}')",
				exc_info=e
			)


	@staticmethod
	def get_conf_file(files: list[str], key: str) -> None | str:
		"""
		Extract the first entry who matches the provided 'key' from the provided list
		Note: this method purposely modifies the input list (removes matched element)
		"""
		for el in files:
			if key in el:
				files.remove(el)
				return el
		return None


	@staticmethod
	def get_conf_files(files: list[str], key: str) -> list[str]:
		"""
		Extract all entries who matches the provided 'key' from the provided list
		Note: this method purposely modifies the input list (removes matched elements)
		"""
		ret = [el for el in files if key in el]
		for el in ret:
			files.remove(el)
		return ret



from contextlib import nullcontext
import io, pytest, yaml, subprocess, tempfile, os
from pathlib import Path
from argparse import Namespace

from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.base.BadConfig import BadConfig
from ampel.config.builder.ConfigChecker import ConfigChecker
from ampel.config.builder.ConfigValidator import ConfigValidator
from ampel.core.UnitLoader import UnitLoader
from ampel.util.mappings import set_by_path

from ampel.test.test_JobCommand import run


def test_build_config():
    tmp_file = os.path.join(tempfile.mkdtemp(), "ampel_conf.yml")
    io.BytesIO(
        subprocess.check_output(
            [
                "ampel",
                "config",
                "build",
                "--distributions",
                "ampel-interface",
                "ampel-core",
                "-out",
                tmp_file,
            ]
        )
    )
    with open(tmp_file, "r") as f:
        config = yaml.safe_load(f)
    assert ConfigValidator(config).validate() == config


def test_ConfigChecker(testing_config, monkeypatch):
    """
    ConfigValidator validates units without calling their __init__ methods
    """
    with open(testing_config) as f:
        config = yaml.safe_load(f)
    # validates as-is
    checker = ConfigChecker(config)
    checker.validate(raise_exc=True)

    # add a processor with side-effects
    class SideEffect(RuntimeError):
        ...

    class SideEffectLadenProcessor(AbsEventUnit):

        required: int

        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            raise SideEffect

        def proceed(self, event_hdlr):
            ...

    config["process"]["t0"]["BadProcess"] = {
        "name": "BadProcess",
        "version": 0,
        "active": True,
        "schedule": "super",
        "processor": {"unit": "SideEffectLadenProcessor", "config": {"required": 1}},
    }
    checker = ConfigChecker(config)

    def get_class_by_name(name, *args, **kwargs):
        if name == "SideEffectLadenProcessor":
            return SideEffectLadenProcessor
        else:
            return UnitLoader.get_class_by_name(checker.loader, name, *args, **kwargs)

    monkeypatch.setattr(checker.loader, "get_class_by_name", get_class_by_name)

    # ConfigChecker attempts to instantiate the unit
    with pytest.raises(SideEffect):
        assert checker.validate(raise_exc=True)

    # ConfigValidator just validates the model
    checker = ConfigValidator(config)
    monkeypatch.setattr(checker.loader, "get_class_by_name", get_class_by_name)
    checker.validate(raise_exc=True)

    # ConfigValidator fails if the config does not satisfy the model
    checker.config["process"]["t0"]["BadProcess"]["processor"]["config"].clear()
    with pytest.raises(BadConfig):
        checker.validate()


@pytest.mark.parametrize("doc", [{"bignumber": 1 << 57}, {1: 2}])
def test_transform_config(doc, tmpdir):
    """Transform preserves objects that are not representable in JSON"""
    infile = Path(tmpdir / "in.yaml")
    outfile = Path(tmpdir / "out.yaml")
    yaml.dump(doc, infile.open("w"))
    assert (
        run(
            [
                "ampel",
                "config",
                "transform",
                "--file",
                str(infile),
                "--out",
                str(outfile),
                "--filter",
                ".",
            ]
        )
        == None
    )
    transformed_doc = yaml.safe_load(outfile.open())
    assert transformed_doc == doc


@pytest.mark.parametrize(
    "patch,result", [({}, None), ({"channel.LONG_CHANNEL.purge": {}}, TypeError)]
)
def test_validate_config(testing_config, tmpdir, patch, result):
    """Validate validates config"""
    tmp_config = tmpdir / "patched_config.yml"
    with testing_config.open() as f:
        config = yaml.safe_load(f)
    for path, item in patch.items():
        set_by_path(config, path, item)
    tmp_config.write_text(yaml.dump(config), "utf-8")
    with pytest.raises(result) if result else nullcontext():
        assert (
            run(
                [
                    "ampel",
                    "config",
                    "validate",
                    "--file",
                    str(tmp_config),
                ]
            )
            == result
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/time/TimeConstraintModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                29.09.2018
# Last Modified Date:  06.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.model.time.TimeDeltaModel import TimeDeltaModel
from ampel.model.time.TimeLastRunModel import TimeLastRunModel
from ampel.model.time.TimeStringModel import TimeStringModel
from ampel.model.time.UnixTimeModel import UnixTimeModel
from ampel.model.time.QueryTimeModel import QueryTimeModel
from ampel.base.AmpelBaseModel import AmpelBaseModel


class TimeConstraintModel(AmpelBaseModel):
	"""
	Examples::
	  
	  TimeConstraintModel(
	      **{
	          "after": {
	              "match_type": "time_delta",
	              "days": -1
	          },
	          "before": {
	              "match_type": "time_string",
	              "dateTimeStr": "21/11/06 16:30",
	              "dateTimeFormat": "%d/%m/%y %H:%M"
	          }
	      }
	  )

	::
	  
	  TimeConstraintModel(
	      **{
	          "after": {
	              "match_type": "time_last_run",
	              "name": "val_test"
	          },
	          "before": {
	              "match_type": "unix_time",
	              "value": 1531306299
	          }
	      }
	  )
	"""

	before: None | TimeDeltaModel | TimeLastRunModel | TimeStringModel | UnixTimeModel = None
	after: None | TimeDeltaModel | TimeLastRunModel | TimeStringModel | UnixTimeModel = None


	def get_query_model(self, **kwargs) -> None | QueryTimeModel:
		"""
		Call this method with db=<instance of AmpelDB>
		if your time constraint is based on TimeLastRunModel
		"""

		if self.before is None and self.after is None:
			return None

		return QueryTimeModel(
			before = self.before.get_timestamp(**kwargs) if self.before else None,
			after = self.after.get_timestamp(**kwargs) if self.after else None
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsDocIngester.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                10.03.2020
# Last Modified Date:  09.10.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any, Generic
from ampel.types import T
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.mongo.update.DBUpdatesBuffer import DBUpdatesBuffer


class AbsDocIngester(Generic[T], AmpelABC, AmpelBaseModel, abstract=True):

	updates_buffer: DBUpdatesBuffer

	@abstractmethod
	def ingest(self, doc: T) -> None:
		...

	def get_stats(self, reset: bool = True) -> None | dict[str, Any]:
		return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsChannelTemplate.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                27.10.2019
# Last Modified Date:  07.04.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.base.decorator import abstractmethod
from ampel.log.AmpelLogger import AmpelLogger
from ampel.model.ChannelModel import ChannelModel
from ampel.config.builder.FirstPassConfig import FirstPassConfig
from ampel.config.builder.AbsConfigTemplate import AbsConfigTemplate


class AbsChannelTemplate(AbsConfigTemplate, ChannelModel, abstract=True):


	def get_channel(self, logger: AmpelLogger) -> dict[str, Any]:
		keys = ChannelModel.get_model_keys()
		return {k: v for k,v in self.__dict__.items() if k in keys}

	@abstractmethod
	def get_processes(self, logger: AmpelLogger, first_pass_config: FirstPassConfig) -> list[dict[str, Any]]:
		...

	def transfer_channel_parameters(self, process: dict[str, Any]) -> dict[str, Any]:
		"""
		Adds channel defined information to the provided process:
		'active, 'distrib', 'source' and 'channel'
		Note: a shallow copy of the input dict is made
		"""
		p = process.copy()

		# Inactivated channels cannot have active processes
		if not self.active:
			p['active'] = False

		# Processes inherit channel distribution/file/name infos
		p['distrib'] = self.distrib
		p['source'] = self.source
		p['channel'] = self.channel

		return p



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/log/handlers/AmpelStreamHandler.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.10.2018
# Last Modified Date:  24.05.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import sys, time
from sys import _getframe
from os.path import basename
from time import strftime
from typing import Literal
from ampel.log.LogFlag import LogFlag
from ampel.log.LightLogRecord import LightLogRecord
from ampel.util.mappings import compare_dict_values
from ampel.util.collections import try_reduce

levels = {1: 'DEBUG', 2: 'VERBOSE', 4: 'INFO', 8: 'SHOUT', 16: 'WARNING', 32: 'ERROR'}

# flake8: noqa: E101
class AmpelStreamHandler:
	"""
	:param int aggregate_interval: logs with similar attributes (log level, ...) are aggregated
	in one document instead of being split into several documents (spares some index RAM).
	*aggregate_interval* is the max interval of time in seconds during which log aggregation occurs.
	Beyond this value, a new log document is created no matter what.
	Note that it impacts the logging time granularity.

	:param density: whether log header (including timestamp) should be displayed on the console:

	- "default" (aggreation break if log-level, class-origin or 'extra' differ between log entries):
	2020-03-03 03:01:23 DistConfigBuilder VERBOSE
	 Loading conf/ampel-contrib-hu/unit.conf from distribution 'ampel-contrib-hu'

	2020-03-03 03:01:24 UnitConfigCollector VERBOSE
	 Adding T0 base unit: DecentFilter
	 Adding T0 base unit: SimpleDecentFilter
	 Adding T0 base unit: XShooterFilter

	- "compact" (aggregates log entries even if log entries originates from different classes):
	2020-03-03 03:01:23 DistConfigBuilder VERBOSE
	 Loading conf/ampel-contrib-hu/unit.conf from distribution 'ampel-contrib-hu'
	 Adding T0 base unit: DecentFilter
	 Adding T0 base unit: SimpleDecentFilter
	 Adding T0 base unit: XShooterFilter

	- "headerless" (Skip the log headers completely (no timestamp, no log level...):
	Loading conf/ampel-contrib-hu/unit.conf from distribution 'ampel-contrib-hu'
	Adding T0 base unit: DecentFilter
	Adding T0 base unit: SimpleDecentFilter
	Adding T0 base unit: XShooterFilter
	"""

	__slots__ = "__dict__", "stream", "prev_record", "timefunc", \
		"prefix", "provenance", "nl", "aggregate_interval"


	def __init__(self,
		std_stream: Literal['stdout', 'stderr'] = 'stderr',
		datefmt: str = "%Y-%m-%d %H:%M:%S",
		level: int = LogFlag.INFO,
		aggregate_interval: float = 1.,
		density: Literal["default", "compact", "compacter", "headerless"] = "default",
		terminator: str = '\n',
		log_sep: str = '\n', # separator between aggregated log entries
		prefix: None | str = None,
		provenance: bool = True
	) -> None:

		self.aggregate_interval = aggregate_interval
		self.dummy_record = LightLogRecord(name='dummy', levelno=1<<10, msg=None)
		self.prev_record: LightLogRecord = self.dummy_record
		self.datefmt = datefmt
		self.prefix = prefix
		self.level = level
		self.nl = terminator
		self.log_sep = log_sep
		self.provenance = provenance
		self.stream = getattr(sys, std_stream)

		if density == "default":
			self.fields_check = ['extra', 'stock', 'channel', 'filename', 'levelno']
		elif density == "compact":
			self.fields_check = ['extra', 'stock', 'channel', 'levelno']
		elif density == "compacter":
			self.log_sep = ''
			self.fields_check = ['extra', 'stock', 'channel']
			self.handle = getattr(self, f"handle_compacter") # type: ignore[assignment]
		elif density == "headerless":
			self.handle = getattr(self, f"handle_headerless") # type: ignore[assignment]


	def handle(self, rec: LightLogRecord) -> None:

		prev_rec = self.prev_record

		# Same flag, date (+- 1 sec), tran_id and chans
		if (
			rec.msg and rec.created - prev_rec.created < self.aggregate_interval and
			compare_dict_values(rec.__dict__, prev_rec.__dict__, self.fields_check) # type: ignore[arg-type]
		):
			self.stream.write(f' {rec.msg}{self.nl}')
		else:
			self.stream.write(f'{self.log_sep}{self.format(rec)}{self.nl}')
			self.prev_record = rec


	def handle_compacter(self, rec: LightLogRecord) -> None:

		prev_rec = self.prev_record

		# Same flag, date (+- 1 sec), tran_id and chans
		if (
			rec.msg and rec.created - prev_rec.created < self.aggregate_interval and
			compare_dict_values(rec.__dict__, prev_rec.__dict__, self.fields_check) # type: ignore[arg-type]
		):
			self.stream.write(f" {rec.msg}{self.nl}")
		else:
			self.stream.write(f"{self.log_sep}{self.format(rec)}{self.nl}")
			self.prev_record = rec


	def handle_headerless(self, rec: LightLogRecord) -> None:
		if rec.msg:
			self.stream.write(rec.msg)
		if rec.extra:
			if rec.msg:
				self.stream.write(f" ")
			self.stream.write(str(rec.extra))
		self.stream.write(self.nl)


	def format(self, record: LightLogRecord) -> str:

		out = strftime(self.datefmt)
		lvl = record.levelno

		if self.prefix:
			out += self.prefix

		# Location
		if lvl & 64:
			out += ' UNIT'
		if lvl & 128:
			out += ' CORE'

		if self.provenance and record.filename:
			if record.filename[0] == '<': # ipython
				print(lvl)
				out += f' {record.filename} {levels[lvl >> 8]}'
			else:
				out += f' {record.filename[:-3]}:{record.lineno} {levels[lvl >> 8]}'
		else:
			out += f' {levels[lvl >> 8]}'

		if record.extra:
			suffix = [f'{k}={record.extra[k]}' for k in record.extra]
		else:
			suffix = []

		if record.stock:
			suffix.append(f's={record.stock}') # type: ignore

		if record.channel:
			suffix.append(f'c={try_reduce(record.channel)}')

		if suffix:
			out += f' [{", ".join(suffix)}]'

		if isinstance(record.msg, str):
			return out + "\n " + record.msg.replace("\n", "\n ") # type: ignore[union-attr]

		return out


	def flush(self) -> None:
		self.stream.flush()


	def break_aggregation(self) -> None:
		self.prev_record = self.dummy_record



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/AbsStockCommand.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                25.03.2021
# Last Modified Date:  25.03.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any, Literal
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.cli.MaybeIntAction import MaybeIntAction
from ampel.cli.LoadJSONAction import LoadJSONAction
from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.mongo.utils import maybe_match_array
from ampel.model.UnitModel import UnitModel
from ampel.model.time.UnixTimeModel import UnixTimeModel
from ampel.model.time.TimeStringModel import TimeStringModel
from ampel.model.time.TimeLastRunModel import TimeLastRunModel
from ampel.model.time.TimeDeltaModel import TimeDeltaModel
from ampel.model.time.TimeConstraintModel import TimeConstraintModel


class AbsStockCommand(AbsCoreCommand, abstract=True):
	"""
	Base class for commands selecting/matching stock(s)
	"""

	@staticmethod
	def get_select_args_help() -> dict[str, str]:

		return {
			# Required
			'config': 'Path to an ampel config file (yaml/json)',
			# Optional
			'secrets': 'Path to a YAML secrets store in sops format',
			'log-profile': 'One of: default, compact, headerless, verbose, debug',
			'id-mapper': 'Convert stock ids using the provided id mapper (ex: ZTFIdMapper)',
			# Selection
			'stock': 'Stock id(s) (OR matched if multi-valued)',
			'channel': 'Channel(s)',
			'created-after-ts': 'Created after unix timestamp',
			'created-after-str': 'Created after date-time iso string',
			'created-after-delta': 'Created after time delta',
			'created-after-process': 'Created after last run of process with name',
			'created-before-ts': 'Created before unix timestamp',
			'created-before-str': 'Created before date-time iso string',
			'created-before-delta': 'Created before time delta',
			'created-before-process': 'Created before last run of process with name',
			'updated-after-ts': 'Updated after unix timestamp',
			'updated-after-str': 'Updated after date-time iso string',
			'updated-after-delta': 'Updated after time delta',
			'updated-after-process': 'Updated after last run of process with name',
			'updated-before-ts': 'Updated before unix timestamp',
			'updated-before-str': 'Updated before date-time iso string',
			'updated-before-delta': 'Updated before time delta',
			'updated-before-process': 'Updated before last run of process with name',
			'custom-match': 'Custom mongodb match as JSON string (ex: {"body.aKey": {"$gt": 1}})',
		}


	def add_selection_args(self, builder: ArgParserBuilder, sub_ops: str = 'all') -> None:

		# Selection args
		builder.add_group('match', 'Stock selection arguments', sub_ops)
		builder.arg('stock', group='match', sub_ops=sub_ops, action=MaybeIntAction, nargs='+')
		builder.xargs(
			group = 'match',
			sub_ops = sub_ops,
			xargs = [
				{'name': 'created-before-str'}, {'name': 'created-before-ts', 'type': int},
				{'name': 'created-before-delta', 'action': LoadJSONAction},
				{'name': 'created-before-process'}
			]
		)
		builder.xargs(
			group = 'match',
			sub_ops = sub_ops,
			xargs = [
				{'name': 'created-after-str'}, {'name': 'created-after-ts', 'type': int},
				{'name': 'created-after-delta', 'action': LoadJSONAction},
				{'name': 'created-after-process'}
			]
		)
		builder.xargs(
			group = 'match',
			sub_ops = sub_ops,
			xargs = [
				{'name': 'updated-before-str'}, {'name': 'updated-before-ts', 'type': int},
				{'name': 'updated-before-delta', 'action': LoadJSONAction},
				{'name': 'updated-before-process'}
			]
		)
		builder.xargs(
			group = 'match',
			sub_ops = sub_ops,
			xargs = [
				{'name': 'updated-after-str'}, {'name': 'updated-after-ts', 'type': int},
				{'name': 'updated-after-delta', 'action': LoadJSONAction},
				{'name': 'updated-after-process'}
			]
		)
		builder.logic_args('channel', descr='Channel', group='match', sub_ops=sub_ops)
		builder.logic_args('with-tag', descr='Tag', group='match', sub_ops=sub_ops)
		builder.logic_args('without-tag', descr='Tag', group='match', sub_ops=sub_ops, excl=True)
		builder.arg(
			'custom-match', group='match', sub_ops=sub_ops,
			metavar='#', action=LoadJSONAction
		)


	def get_tag(self, args: dict[str, Any]) -> None | dict[Literal['with', 'without'], dict]:

		tag: None | dict[Literal['with', 'without'], dict] = None
		if args.get('with_tag'):
			tag = {'with': args['with_tag']}
		if args.get('without_tag'):
			if tag is None:
				tag = {}
			tag['without'] = args['without_tag']
		return tag


	def build_select_model(self, args: dict[str, Any]) -> UnitModel:

		conf = {
			'created': self.get_time_model('created', args),
			'updated': self.get_time_model('updated', args),
			'channel': args['channel'],
			'custom': args['custom_match']
		}

		if args.get('tag'):
			conf['tag'] = self.get_tag(args)

		if (stock := args.get('stock')):
			conf['custom'] = { # Hmmm, unsure about the necessity of 'custom' there, check later
				'stock': stock if isinstance(stock, (int, bytes, str))
					else maybe_match_array(stock)
			}

		return UnitModel(unit='T3StockSelector', config=conf)


	def get_time_model(self, prefix: str, args: dict[str, Any]) -> TimeConstraintModel:

		d: dict[str, Any] = {'after': None, 'before': None}

		for when in ('after', 'before'):
			if args.get(x := f'{prefix}_{when}_ts'):
				d[when] = UnixTimeModel(match_type='unix_time', value=args[x])
			elif args.get(x := f'{prefix}_{when}_str'):
				d[when] = TimeStringModel(
					match_type='time_string',
					dateTimeStr=args[x],
					dateTimeFormat='%Y%m%dT%H%M%S'
				)
			elif args.get(x := f'{prefix}_{when}_delta'):
				d[when] = TimeDeltaModel(match_type='time_delta', **args[x])
			elif args.get(x := f'{prefix}_{when}_process'):
				d[when] = TimeLastRunModel(match_type='time_last_run', process_name=args[x])

		return TimeConstraintModel(**d)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/test/test_ProcessMorpher.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                Unspecified
# Last Modified Date:  23.04.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import pytest, yaml
from ampel.config.builder.ProcessMorpher import ProcessMorpher
from ampel.config.builder.DisplayOptions import DisplayOptions
from ampel.config.collector.T02ConfigCollector import T02ConfigCollector


@pytest.fixture
def first_pass_config(testing_config):
    with open(testing_config) as f:
        return yaml.safe_load(f)


@pytest.fixture
def config_collector(first_pass_config, ampel_logger):
    first_pass_config["confid"] = T02ConfigCollector(
        conf_section = "confid",
		options = DisplayOptions(),
		logger = ampel_logger
    )
    return first_pass_config


@pytest.mark.parametrize(
    "config,hashed_config",
    [
        ({}, {"test_parameter": 1}),
        ({"test_parameter": 1}, {"test_parameter": 1}),
        ({"test_parameter": 2}, {"test_parameter": 2}),
    ],
)
def test_hash_t2_config(config_collector, config, hashed_config, ampel_logger):

    m = ProcessMorpher(
        {
            "active": True,
            "channel": "HU_RANDOM",
            "controller": {"unit": "ZTFAlertStreamController"},
            "distrib": "ampel-hu-astro",
            "name": "HU_RANDOM|T0|ztf_uw_private",
            "processor": {
                "config": {
                    "compiler_opts": None,
                    "directives": [
                        {
                            "channel": "HU_RANDOM",
                            "filter": {
                                "config": {"passing_rate": 0.0001},
                                "unit": "RandFilter",
                            },
                            "ingest": {
                                "point_t2": [{"unit": "DemoPointT2Unit", "config": config}]
                            },
                        }
                    ],
                    "shaper": {"unit": "ZiDataPointShaper"},
                    "supplier": {
                        "config": {
                            "loader": {
                                "config": {
                                    "bootstrap": "partnership.alerts.ztf.uw.edu:9092",
                                    "group_name": "ampel-v0.7.1",
                                    "stream": "ztf_uw_private",
                                    "timeout": 3600,
                                },
                                "unit": "UWAlertLoader",
                            }
                        },
                        "unit": "ZiAlertSupplier",
                    },
                },
                "unit": "AlertConsumer",
            },
            "schedule": ["super"],
            "source": "/Users/jakob/Documents/ZTF/Ampel-v0.8/Ampel-HU-astro/conf/ampel-hu-astro/channel/HU_RANDOM.yml",
            "tier": 0,
        },
        templates=[],
        logger=ampel_logger,
    )
    m.apply_template(config_collector)

    m.hash_t2_config(config_collector)
    assert (
        len(configs := list(config_collector["confid"].values())) == 1
    ), "config hash emitted"
    assert configs[0] == hashed_config, "validated config appears in output"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/core/DataLoader.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                13.01.2018
# Last Modified Date:  22.04.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from bson.codec_options import CodecOptions
from typing import Literal
from collections.abc import Iterable, Iterator

from ampel.types import StockId, ChannelId, StrictIterable, Tag
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.OneOf import OneOf
from ampel.model.t3.LoaderDirective import LoaderDirective
from ampel.mongo.view.FrozenValuesDict import FrozenValuesDict
from ampel.mongo.query.general import build_general_query
from ampel.log.utils import safe_query_dict
from ampel.log.AmpelLogger import AmpelLogger
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.util.collections import ampel_iter
from ampel.core.AmpelContext import AmpelContext
from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry

# Monitoring counters
stat_db_loads = AmpelMetricsRegistry.counter(
	"loads", "Number of documents loaded",
	subsystem="db", labelnames=("col",)
)

class DataLoader:

	def __init__(self, ctx: AmpelContext) -> None:
		self.ctx = ctx

	def load(self,
		stock_ids: StockId | Iterator[StockId] | StrictIterable[StockId],
		directives: Iterable[LoaderDirective],
		channel: None | ChannelId | AllOf[ChannelId] | AnyOf[ChannelId] | OneOf[ChannelId] = None,
		tag: None | dict[Literal['with', 'without'], Tag | dict | AllOf[Tag] | AnyOf[Tag] | OneOf[Tag]] = None,
		auto_project: bool = True,
		codec_options: None | CodecOptions = CodecOptions(document_class=FrozenValuesDict),
		logger: None | AmpelLogger = None
	) -> Iterable[AmpelBuffer]:
		"""
		:param directives: see LoaderDirective docstrings for more information.  Notes:
		- T3 unit configurations can use aliased directives. For example, the alias COMPOUND
		is translated into the directive LoaderDirective(col="t1")
		- LoaderDirective can contain "query_complement", which can be used among other thing to further
		sub-select t2 documents or select given states (such as the latest state)

		:param tag: If specified, query selection critera will apply to all directives.

		:param channel: If specified, query selection critera will apply to all directives (except t0).
		None means all channels are considered (no criterium).

		:param auto_project: each LoaderDirective is associated with a db collection with is itself associated with
		a default "content" typed dict. By default, we request a projection that projects only the fields defined
		in those TypedDict. Custom/admin fields would thus not be retrieved.
		Set this setting to False if it is not the whished behavior.
		"""

		col_set = {directive.col for directive in directives}

		# Note: the following operation will consume
		# stock_ids if it is an Iterator/Cursor
		register: dict[StockId, AmpelBuffer] = {
			stock_id: AmpelBuffer(
				id = stock_id,
				stock = None if "stock" in col_set else None,
				t0 = [] if "t0" in col_set else None,
				t1 = [] if "t1" in col_set else None,
				t2 = [] if "t2" in col_set else None
			) for stock_id in ampel_iter(stock_ids)
		}

		for directive in directives:

			query = build_general_query(
				stock=register.keys(), channel=channel, tag=tag
			)

			#if directive.col == "stock":
			#	query['_id'] = query.pop("stock")
			#elif directive.col == "t0" and channel:
			#	query.pop('channel')

			# query 'stock' parameter primes over query complements
			if directive.query_complement:
				query = directive.query_complement | query

			if logger and logger.verbose > 1: # log query parameters
				logger.debug(
					None, extra={
						'col': directive.col,
						'query': safe_query_dict(query, dict_key=None)
					}
				)

			# Retrieve pymongo cursor
			col = self.ctx.db.get_collection(directive.col)

			if codec_options:
				col = col.database.get_collection(col.name, codec_options=codec_options)

			# Note: codec_options freezes structures in dicts with depth level > 1
			cursor = col.find(
				filter = query,
				projection = {
					k: 1 for k in directive.model.__annotations__.keys()
				} if auto_project else None
			)

			inc = stat_db_loads.labels(directive.col).inc

			if directive.col == "t1":
				count = 0
				for count, res in enumerate(cursor, 1):
					register[res['stock']][directive.col].append(res) # type: ignore[union-attr]
				inc(count)

			elif directive.col == "stock":
				count = 0
				for count, res in enumerate(cursor, 1):
					register[res['stock']]['stock'] = res
				inc(count)

			# Datapoints are potentially channel-less and can be associated with multiple stocks
			elif directive.col == "t0":

				count = 0
				for count, res in enumerate(cursor, 1):

					# Upper limits can be attached to multiple transients
					for sid in res['stock']:

						# Some of which might not match with our query
						if sid not in register:
							continue

						# Add datapoints to snapdata (no need to recursive_freeze
						# since dict elements with level > 1 are frozen due to codec_options
						register[sid]['t0'].append(res) # type: ignore[union-attr]
				inc(count)

			elif directive.col == "t2":

				# the entire data will need to fit in memory anyway
				res = list(cursor)

				# whether to replace init config integer hash with 'resolved' config dict
				if directive.resolve_config:

					config_keys = self.ctx.config.get('t2.config_keys', dict)
					if not config_keys:
						raise ValueError("Cannot load t2 configs")

					for el in res:
						#: init config integer hash with 'resolved' config dict
						dict.__setitem__(el, 'config', el[config_keys[el['config']]]) # type: ignore[index]

				inc(len(res))

				if directive.excluding_query:

					sids = set(register.keys())
					for el in res:
						if el['stock'] in sids:
							sids.remove(el['stock'])
					if sids:
						for k in sids:
							del register[k]

				for el in res:
					register[el['stock']]['t2'].append(el) # type: ignore[union-attr]

			else:
				raise ValueError(
					f"Unrecognized LoaderDirective: {directive.dict()}"
				)

		if logger and logger.verbose:
			s = f"Unique ids: {len(register)}"
			for col_name in (col_set - set(["stock"])):
				s += f", {col_name}: "
				s += str(sum([1 for k in register for el in register[k][col_name]])) # type: ignore
			logger.info(s)

		return register.values()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/ingest/CompilerOptions.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                08.05.2021
# Last Modified Date:  15.05.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.model.aux.AuxAliasableModel import AuxAliasableModel


empty: dict[str, Any] = {}

class CompilerOptions(AuxAliasableModel):
	"""
	Will be merged with the options set by say IngestionHandlers (these will have priority).
	Allows for example to set default tags for given documents or to define a custom AbsIdMapper
	subclass for the stock compiler.
	"""

	t0: dict[str, Any] = empty
	t1: dict[str, Any] = empty
	state_t2: dict[str, Any] = empty
	point_t2: dict[str, Any] = empty
	stock_t2: dict[str, Any] = empty
	stock: dict[str, Any] = empty



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/utils.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                24.03.2021
# Last Modified Date:  27.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import os
from typing import Any, Literal
from ampel.core.AmpelDB import AmpelDB
from ampel.secret.AmpelVault import AmpelVault
from ampel.config.AmpelConfig import AmpelConfig
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.abstract.AbsIdMapper import AbsIdMapper
from ampel.abstract.AbsSecretProvider import AbsSecretProvider
from ampel.util.collections import check_seq_inner_type


def get_vault(args: dict[str, Any]) -> None | AmpelVault:
	vault = None
	if args.get('secrets'):
		from ampel.secret.DictSecretProvider import DictSecretProvider
		from ampel.secret.DirSecretProvider import DirSecretProvider
		if os.path.isdir(args['secrets']):
			provider: AbsSecretProvider = DirSecretProvider(args['secrets'])
		else:
			provider = DictSecretProvider.load(args['secrets'])
		vault = AmpelVault([provider])
	return vault


def get_db(
	config: AmpelConfig,
	vault: None | AmpelVault = None,
	require_existing_db: bool | str = True,
	one_db: bool | Literal['auto'] = False,
) -> AmpelDB:

	try:
		return AmpelDB.new(
			config,
			vault,
			require_exists = require_existing_db,
			one_db = one_db
		)
	except Exception as e:
		if "Databases with prefix" in str(e):
			s = "Databases with prefix " + (
				require_existing_db if isinstance(require_existing_db, str) else
				config.get('mongo.prefix', str, raise_exc=True)
			) + " do not exist"
			raise SystemExit("\n" + "="*len(s) + "\n" + s + "\n" + "="*len(s) + "\n")
		raise e


def _maybe_int(stringy):
	try:
		return int(stringy)
	except Exception:
		return stringy


def maybe_load_idmapper(args: dict[str, Any]) -> None:
	"""
	Replaces the string id defined in args['id_mapper'] with an instance of the requested id mapper.
	Replaces potential string stock ids with their ampel ids.
	"""

	args['id_mapper'] = AuxUnitRegister.get_aux_class(
		args['id_mapper'], sub_type=AbsIdMapper
	)() if args['id_mapper'] else None

	if not args['id_mapper']:
		return

	if isinstance(args['stock'], str):
		if "," in args['stock']:
			s = [int(x) if (x := el.strip()).isdigit() else x for el in args['stock'].split(",")]
			args['stock'] = args['id_mapper'].to_ampel_id(s)
		else:
			args['stock'] = args['id_mapper'].to_ampel_id(args['stock'])

	elif check_seq_inner_type(args['stock'], str):
		args['stock'] = args['id_mapper'].to_ampel_id(args['stock']) # type: ignore



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/template/ChannelWithProcsTemplate.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.10.2019
# Last Modified Date:  05.01.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.log.AmpelLogger import AmpelLogger
from typing import Any
from ampel.config.builder.FirstPassConfig import FirstPassConfig
from ampel.abstract.AbsChannelTemplate import AbsChannelTemplate
from ampel.model.ChannelModel import ChannelModel

class ChannelWithProcsTemplate(AbsChannelTemplate):
	""" Convenience class allowing channel definitions to include processes.  """

	# Note: not using list[ProcessModel] on purpose since embedded processes
	# might need template processing as well
	process: list[dict[str, Any]]

	def get_channel(self, logger: AmpelLogger) -> dict[str, Any]:
		return self.dict(include=ChannelModel.get_model_keys())

	def get_processes(self, logger: AmpelLogger, first_pass_config: FirstPassConfig) -> list[dict[str, Any]]:

		# Note: not enforcing channel selection for t3 processes
		# as these could require template processing first

		return [
			self.transfer_channel_parameters(p)
			for p in self.process
		]
